---
--    Copyright (c) 2016 by SAS Institute Inc., Cary, NC, USA.
-- 
--- REST service manipulation routines
--  you first need to set the url using set_url(). 
--  Set the authentication method with set_auth()
--  There is a table called M.htttp_codes that you can use to look up the return codes
--  returned from request()
--  @usage 
-- EXAMPLE: Using the generic request function
--   
--   local rest       = require 'risk.stress.rest_service'
--   local json       = require 'risk.stress.json'
--   local utils      = require 'risk.stress.rest_utils'
--
--   URL = "http://rdcesx06011.race.sas.com:7980/SASRiskandFinanceWorkbench/rest/"
--
--   local pass = rest.set_url(URL)
--   print(rest.url)
--   local pass = rest.set_auth("token")
--   print(rest.auth)
--
--   pass, json, code = rest.request("get","projects")
--   print(pass)
--   print(json)
--   print(code)
--   msg = rest.http_codes[code]
--   print(msg)
--
--[[ EXAMPLE USING AUTHENTICATION ON THE MIDTIER:
      filename LUAPATH  ( "C:\cvs\");
      local rest = require 'risk.stress.rest_service'
      local JSON = require 'risk.stress.JSON'
      local pass,json_out;
      JSON.strictTypes = true
      pass = rest.set_url('http://v64-d7b797-01.na.sas.com:7980/SASRiskandFinanceWorkbench/rest/')
      print(pass)
      json_out, pass = rest.request_json_tokenauth( "get", "?" )
      Print the Results
      print(json_out, pass)
      local l = JSON:decode(json_out  )
      print(table.tostring(l))
]]--
--[[ EXAMPLE USING TICKET GRANTING TICKET:
      filename LUAPATH  ( "C:\cvs\");
      local rest = require 'risk.stress.rest_service'
      local JSON = require 'risk.stress.JSON'
      local pass,json_out;
      JSON.strictTypes = true
      pass = rest.set_url('http://v64-d7b797-01.na.sas.com:7980/SASRiskandFinanceWorkbench/rest/')
      pass = rest.set_rest_ticket_granting_ticket_cas( "sasdemo", "Go4thsas")
      json_out, pass =  rest.request_json_tgt( "get", "?" )
      local l = JSON:decode(json_out  )
      print(table.tostring(l))
]]--
-- @module rest_service
local M={}
require 'risk.stress.global_init'
local errors=require 'risk.stress.errors'
local utils=require 'risk.stress.rest_utils'
local string_utils=require 'risk.stress.string_utils'
local json=require 'risk.stress.json'
local metadata=require 'risk.stress.metadata'
-- To be set with a call to set_url
M.url=""
--To be set with a call to set_auth. Default is token
M.auth="token"
M.rest_service_ticket=""
M.rest_ticket_granting_ticket=""
local temp_folder = sas.getoption('work')
local basic_auth = false
local basic_auth_header = ""
--- Set the URL and filenames for http requests.
-- @param in_url  - a string containing the url
function M.set_url(in_url)
   if in_url and M.url == in_url then return true end -- nothing to do
   M.url = in_url
   Debug("URL set to " .. M.url)
   local file_sep = utils.get_file_sep();
   -- Assign filerefs for headers and footers now to be used later so
   -- each individual function doesn;t need to do this. 
   sas.submit([[
      filename hin  "@temp_folder@@file_sep@_hin.txt";
      filename hout "@temp_folder@@file_sep@_hout.txt";
      filename in   "@temp_folder@@file_sep@_in.txt";
      filename out  "@temp_folder@@file_sep@_out.txt";
   ]], {temp_folder = temp_folder, file_sep = file_sep })   
   return true
end
---Sets the authentication method. Valid options are token, ticket, and basic
--- If you use ticket generating ticket, you need to call set_rest_ticket_granting_ticket_cas before making calls to the api
-- @param auth_method - the method to use.
-- @return pass - boolean whether or not the value was set.
function M.set_auth(auth_method)
   local method = sas.lowcase(auth_method)
   if method == "token" then
      M.auth = "token"
   elseif method == "ticket" then
      M.auth = "ticket"
   else
--       sas.print("%1z %s is not a valid authentication method. Valid options are 'token' and 'ticket'",auth_method)
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_set_auth_1', 'NOQUOTE'),auth_method)
--       sas.print("%2zThe previous  authentication method, %s will be used",M.auth)
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_set_auth_2', 'NOQUOTE'),M.auth)
      return false
   end
   return true
end
--- Submit a request to M.url using the M.auth authentication method
--- Output is captured in the filerefs out and hout
--- For header_in and body, you can pass a table instead of text. If you do pass a table, header_in[1] and body[1] 
--- are assumed to be filerefs.
-- @param action - string, GET, POST, PUT, etc.
-- @param request - the portion of the URL that follows M.url
-- @param body[OPT] - text for the body. If nil, then no body is passed
-- @param header_in[OPT] - text for the input header. If nil, then no header is passed (proc http default is used)
-- @param content_type[OPT] - This is the content type. You can pass this in with the header but proc http has a separate argument for this
-- @param out[OPT] - The fileref to write the output to - defaults to 'out' - If not specified the body will be returned from the call as a string.
-- @return pass - whether or not the HTTP return code was in the 200's (200 OK, 201 CREATED, etc)
-- @return json - The output from the call as text if no fileref is specified
-- @return code - the actual http return code
-- @usage
-- pass, json = rest.request( "get", "config/hierarchies")
-- local pass, json = rest.request( "get", "ruleFlows/".. ruleFlowId.."/deployments", nil , 
--                                  "Accept:application/vnd.sas.businessrules.deployments+json;version=1")
function M.request( action, request, body, header_in, content_type, out )
   if M.auth == "token" then 
      local body_arg, header_arg, ct_arg, out_arg
      
      -- Set the 'in' argument to proc http
      if not body then 
         body_arg = " " 
      else
         if type(body) == 'table' then
            body_arg = "in = "..body[1]
         else 
            utils.write_fileref("in",body)
            body_arg = "in = in"
         end
      end
      -- Set the 'headerin' argument to proc http
      if not header_in then 
         header_arg = " " 
      else
         if type(header_in) == 'table' then
            header_arg = "headerin = "..header_in[1]
         else 
            utils.write_fileref("hin",header_in)
            header_arg = "headerin = hin"
         end
      end      
      -- Set the content type and out arguments
      if not content_type then ct_arg = " " else ct_arg = "ct = '"..content_type.."'" end
      if not out then out_arg = "out=out" else out_arg = 'out='..out end
      
     -- If in debug, print the body
      if is_debug_enabled() and action:lower() == "post" then
         local bdy = "in"
         if type(body) == 'table' then bdy = body[1] end
         print(utils.read_fileref(bdy))
      end
 
    sas.submit([[
       proc http
          http_tokenauth
          clear_cache       
          url="%nrstr(@url@@request@)"  
          @header_arg@  
          headerout=hout 
          @body_arg@ 
          @out_arg@
          @ct_arg@
          method='@action@';
       run;
     ]], {url= M.url})
   
   elseif M.auth == "ticket" then
      -- Not supported yet
--       sas.print("%1z Ticket generating tickets for authentication should be handled with request_json_tgt().")
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_request_1', 'NOQUOTE'))
      return false, nil, nil
   else 
      local auth_method
      if not M.auth then auth_method = "nil" else auth_method = M.auth end
--       sas.print("%1z The authentication method (%s) isn't recognized.",auth_method)
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_request_2', 'NOQUOTE'),auth_method)
      return false, nil, nil
   end
   
   --Grab the http return code
   local header_string = utils.read_fileref("hout")
   local header_line_one = sas.scan(header_string,1,'\n\r')
   local code = tonumber(sas.scan(header_line_one,2,' '))
   local pass  = (code >= 200 and code < 300)
   
   -- decide whether to return the body as text.
   local return_body
   if out then
      return_body = "See body in fileref "..out
   else
      return_body = utils.read_fileref('out')
   end
   
   return pass, return_body, code
end


--- Similar to M.request() but uses Host Authentication (basic auth) instead of relying on Token Auth.
function M.request2( action, request, body, header_in, content_type, out, userid, password )

   if M.auth == "token" then 
      local body_arg, header_arg, ct_arg, out_arg
      
      -- Set the 'in' argument to proc http
      if not body then 
         body_arg = " " 
      else
         if type(body) == 'table' then
            body_arg = "in = "..body[1]
         else 
            utils.write_fileref("in",body)
            body_arg = "in = in"
         end
      end
      -- Set the 'headerin' argument to proc http
      if not header_in then 
         header_arg = " " 
      else
         if type(header_in) == 'table' then
            header_arg = "headerin = "..header_in[1]
         else 
            utils.write_fileref("hin",header_in)
            header_arg = "headerin = hin"
         end
      end      
      -- Set the content type and out arguments
      if not content_type then ct_arg = " " else ct_arg = "ct = '"..content_type.."'" end
      if not out then out_arg = "out=out" else out_arg = 'out='..out end
      
     -- If in debug, print the body
      if is_debug_enabled() and action:lower() == "post" then
         local bdy = "in"
         if type(body) == 'table' then bdy = body[1] end
         print(utils.read_fileref(bdy))
      end
 
      -- Use userid/password from mid-tier if not supplied, otherwise escape the given values
      -- to pass into the double quoted string submitted to SAS
      if not userid or userid:len()==0 then
         userid = '%sst_uid'
      else
         userid = string_utils.escape_sas_string_value(userid)
      end
      if not password or password:len()==0 then
         password = '%sst_pwd'
      else
         password = string_utils.escape_sas_string_value(password)
      end

    sas.submit([[

       proc http
          AUTH_BASIC webusername="@userid@" webpassword="@password@"
          clear_cache       
          url="%nrstr(@url@@request@)"  
          @header_arg@  
          headerout=hout 
          @body_arg@ 
          @out_arg@
          @ct_arg@
          method='@action@';
       run;
     ]], {url= M.url})
   
   elseif M.auth == "ticket" then
      -- Not supported yet
--       sas.print("%1z Ticket generating tickets for authentication should be handled with request_json_tgt().")
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_request_1', 'NOQUOTE'))
      return false, nil, nil
   else 
      local auth_method
      if not M.auth then auth_method = "nil" else auth_method = M.auth end
--       sas.print("%1z The authentication method (%s) isn't recognized.",auth_method)
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_request_2', 'NOQUOTE'),auth_method)
      return false, nil, nil
   end
   
   --Grab the http return code
   local header_string = utils.read_fileref("hout")
   local header_line_one = sas.scan(header_string,1,'\n\r')
   local code = tonumber(sas.scan(header_line_one,2,' '))
   local pass  = (code >= 200 and code < 300)
   
   -- decide whether to return the body as text.
   local return_body
   if out then
      return_body = "See body in fileref "..out
   else
      return_body = utils.read_fileref('out')
   end
   
   return pass, return_body, code
end


--- Get a CAS REST service ticket. This returns 'true' if successful and sets the value
--- of 'rest_ticket_granting_ticket' to be used by get_service_ticket.
-- @param username  - a string containing the user name
-- @param password  - a string containing the user password
function M.set_rest_ticket_granting_ticket_cas( username, password )
    
    local url2 = string.match(M.url, "^(http://.+/SAS).*$") .. 'Logon/v1/tickets'
    
    utils.write_fileref("hin", 'accept:text/plain')
    utils.write_fileref_serialized("in", 'username='..username..'&password='..password)
    
    -- Get a ticket granting ticket - TGT 
    sas.submit([[
       proc http 
          url=%str("@url2@")
          headerin=hin 
          headerout=hout 
          in=in method='post' 
          ct='text/plain';
       run;
     ]])
     
     local ret = utils.read_fileref('hout')
     
     -- find the location keyword
     indx = sas.index(ret, 'Location: ')
     if indx  < 1 then return nil end
     ret = sas.substr(ret, indx+10)
     indx = sas.index(ret, '\n')
     M.rest_ticket_granting_ticket = sas.trim( sas.substr(ret, 1, indx-1))
     
     Debug("-------------------------------------------")
     Debug("DEBUG FOR TICKET GRANTING TICKET REQUEST:")
     Debug("URL: " .. url2)
     local hi_str = utils.read_fileref('hin')
     Debug("Header In:" .. hi_str)
     local ho_str = utils.read_fileref('hout')
     Debug("Header Out:" .. ho_str)
     local in_str = utils.read_fileref('in')
     Debug("Body In:" .. in_str)
     local out_str = utils.read_fileref('out')
     Debug("Body Out:" .. out_str)
     Debug("Ticket:" .. M.rest_service_ticket )
     Debug("-------------------------------------------") 
     if sas.index(M.rest_ticket_granting_ticket, 'TGT-') == 0 then 
--         sas.print("%1zFailed to get ticket granting ticket for REST API.")
        sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_set_rest_ticket_granting_ticket_cas_1', 'NOQUOTE'))
        return nil
     end
     -- This is a required initialization call
     M.request_json_tgt( "get", "?" )
       
    return true
end
--- GET operation on a REST service. Must have a ticket granting ticket first. 
--- This is called by request_json_tgt() for you so you shouldn't have to call this.
-- @param request - the service url for which the service ticket is requested
-- @return pass -  boolean whether the header said OK
function M.get_service_ticket( request )
     Debug("Getting service ticket")
     if  basic_auth then 
        return true 
     end
     -- get the service ticket -> ST
     local service_url = 'service='..utils.urlencode(M.url..request)
     utils.write_fileref("in", service_url )
     sas.submit( [[
       proc http 
          url=%nrstr("@rest_ticket_granting_ticket@") 
          headerout=hout
          in=in
          out=out
          method='post' 
          ct='text/plain';
        run;
     ]],  {rest_ticket_granting_ticket = M.rest_ticket_granting_ticket})
     M.rest_service_ticket = utils.read_fileref('out')
     
     Debug("-------------------------------------------")
     Debug("DEBUG FOR SERVICE TICKET REQUEST:")
     Debug("Header Out:")
     local h_str = utils.read_fileref('hout')
     Debug(h_str)
     Debug("Body In:")
     local in_str = utils.read_fileref('in')
     Debug(in_str)
     Debug("Body Out:")
     local out_str = utils.read_fileref('out')
     Debug(out_str)
     Debug("Ticket:")
     Debug( M.rest_service_ticket )
     Debug("-------------------------------------------")
  
     if sas.index(M.rest_service_ticket, 'ST-') == 0 then 
--         sas.print("%1zFailed to get service ticket for REST API.")
        sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_get_service_ticket_1', 'NOQUOTE'))
        return false
     end 
     
      
      return true;
 end
 
--- GET, POST, DELETE, or PUT operation on a REST service using a tgt for auth. 
-- Must have a ticket granting ticket first. 
-- @param action - get, post, delete, or put
-- @param request - the service url for which the service ticket is requested
-- @param body  - for post and put, their will be a body
-- @param header_in - Text to pass as the header in to the request. Default is 'Accept: application/json, text/plain, * / *'
-- @return json, pass - string and boolean whether the header said OK
function M.request_json_tgt( action, request, body, header_in )
 
    if not basic_auth and M.rest_ticket_granting_ticket == ""  then 
--       sas.print("%1zNeed a ticket generating ticket first.")
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_request_json_tgt_1', 'NOQUOTE'))
      return nil
    end 
   
    -- Get a CAS service ticket
    if not M.get_service_ticket( request ) then 
       return nil, false 
    end
     
    -- HEADER: 
    if header_in == nil then
       utils.write_fileref("hin", 'Accept: application/json, text/plain, * / *' )
    else
       utils.write_fileref("hin", header_in )
    end
    --BODY:
    body = body or ""
    utils.write_fileref("in", body )
 
    local ticket_format
    if sas.index(request, '?') ~= 0 then
        ticket_format = '&'
    else
        ticket_format = '?'
    end
    sas.submit([[
       proc http  
          url="%nrstr(@url@@request@@ticket_format@ticket=@rest_service_ticket@)"  
          headerin=hin  
          headerout=hout 
          in=in 
          out=out
          method='@action@'  ;
       run;
     ]], {rest_service_ticket = M.rest_service_ticket, request = request, url= M.url, action=action, body=body, ticket_format=ticket_format})
     
    local h_str = utils.read_fileref('hout')
    local out_str = utils.read_fileref('out')
    Debug("-------------------------------------------")
    Debug("DEBUG FOR REQUEST: " .. M.url .. request .. ticket_format .. "ticket=" .. M.rest_service_ticket)
    Debug("HEADER:")
    Debug(h_str)
    Debug("BODY:")
    Debug(out_str)
    Debug("-------------------------------------------")
    
    local pass = false
    if sas.index(h_str, '200 OK') ~= 0 then 
       pass = true
    end
    return out_str,pass
end
--Alias the legacy tokenauth request function 
M.request_json_tokenauth = M.request
M.http_codes={}
M.http_codes[100]='Continue'
M.http_codes[101]='Switching Protocols'
M.http_codes[102]='Processing (WebDAV; RFC 2518)'
M.http_codes[200]='OK'
M.http_codes[201]='Created'
M.http_codes[202]='Accepted'
M.http_codes[203]='Non-Authoritative Information (since HTTP/1.1)'
M.http_codes[204]='No Content'
M.http_codes[205]='Reset Content'
M.http_codes[206]='Partial Content (RFC 7233)'
M.http_codes[207]='Multi-Status (WebDAV; RFC 4918)'
M.http_codes[208]='Already Reported (WebDAV; RFC 5842)'
M.http_codes[226]='IM Used (RFC 3229)'
M.http_codes[300]='Multiple Choices'
M.http_codes[301]='Moved Permanently'
M.http_codes[302]='Found'
M.http_codes[303]='See Other (since HTTP/1.1)'
M.http_codes[304]='Not Modified (RFC 7232)'
M.http_codes[305]='Use Proxy (since HTTP/1.1)'
M.http_codes[306]='Switch Proxy'
M.http_codes[307]='Temporary Redirect (since HTTP/1.1)'
M.http_codes[308]='Permanent Redirect (RFC 7538)'
M.http_codes[400]='Bad Request'
M.http_codes[401]='Unauthorized (RFC 7235)'
M.http_codes[402]='Payment Required'
M.http_codes[403]='Forbidden'
M.http_codes[404]='Not Found'
M.http_codes[405]='Method Not Allowed'
M.http_codes[406]='Not Acceptable'
M.http_codes[407]='Proxy Authentication Required (RFC 7235)'
M.http_codes[408]='Request Timeout'
M.http_codes[409]='Conflict'
M.http_codes[410]='Gone'
M.http_codes[411]='Length Required'
M.http_codes[412]='Precondition Failed (RFC 7232)'
M.http_codes[413]='Payload Too Large (RFC 7231)'
M.http_codes[414]='URI Too Long (RFC 7231)'
M.http_codes[415]='Unsupported Media Type'
M.http_codes[416]='Range Not Satisfiable (RFC 7233)'
M.http_codes[417]='Expectation Failed'
M.http_codes[418]="I'm a teapot (RFC 2324)"
M.http_codes[421]='Misdirected Request (RFC 7540)'
M.http_codes[422]='Unprocessable Entity (WebDAV; RFC 4918)'
M.http_codes[423]='Locked (WebDAV; RFC 4918)'
M.http_codes[424]='Failed Dependency (WebDAV; RFC 4918)'
M.http_codes[426]='Upgrade Required'
M.http_codes[428]='Precondition Required (RFC 6585)'
M.http_codes[429]='Too Many Requests (RFC 6585)'
M.http_codes[431]='Request Header Fields Too Large (RFC 6585)'
M.http_codes[451]='Unavailable For Legal Reasons'
M.http_codes[500]='Internal Server Error'
M.http_codes[501]='Not Implemented'
M.http_codes[502]='Bad Gateway'
M.http_codes[503]='Service Unavailable'
M.http_codes[504]='Gateway Timeout'
M.http_codes[505]='HTTP Version Not Supported'
M.http_codes[506]='Variant Also Negotiates (RFC 2295)'
M.http_codes[507]='Insufficient Storage (WebDAV; RFC 4918)'
M.http_codes[508]='Loop Detected (WebDAV; RFC 5842)'
M.http_codes[510]='Not Extended (RFC 2774)'
M.http_codes[511]='Network Authentication Required (RFC 6585)'
M.http_codes[103]='Checkpoint'
M.http_codes[420]='Method Failure (Spring Framework)'
M.http_codes[450]='Blocked by Windows Parental Controls (Microsoft)'
M.http_codes[499]='Token Required or Request has been forbidden by antivirus]'
M.http_codes[509]='Bandwidth Limit Exceeded (Apache Web Server/cPanel)'
M.http_codes[530]='Site is frozen'
M.http_codes[440]='Login Timeout'
M.http_codes[449]='Retry With'
M.http_codes[451]='Redirect'
M.http_codes[444]='No Response'
M.http_codes[495]='SSL Certificate Error'
M.http_codes[496]='SSL Certificate Required'
M.http_codes[497]='HTTP Request Sent to HTTPS Port'
M.http_codes[499]='Client Closed Request'
M.http_codes[520]='Unknown Error'
M.http_codes[521]='Web Server Is Down'
M.http_codes[522]='Connection Timed Out'
M.http_codes[523]='Origin Is Unreachable'
M.http_codes[524]='A Timeout Occurred'
M.http_codes[525]='SSL Handshake Failed'
M.http_codes[526]='Invalid SSL Certificate'
-- ******************************************************
--- THESE ARE NOT USED AND SO ARE NOT WELL TESTED
-- ******************************************************
--- Get a basic REST service ticket. This returns a string with the service ticket
-- @param username  - a string containing the user name
-- @param password  - a string containing the user password
function M.set_rest_service_ticket_basic( username, password )
    
    Debug( temp_folder )
    sas.submit([[
       filename hin  "@temp_folder@/_hin.txt";
       filename hout "@temp_folder@/_hout.txt";
       filename in   "@temp_folder@/_in.txt";
       filename out  "@temp_folder@/_out.txt";
    ]], {temp_folder = temp_folder })
    
    basic_auth_header = 'authorization:Basic ' .. sasz.b64encode(username..':'..password)
    
    utils.write_fileref("hin", basic_auth_header .. '\naccept:application/xml')
    
    basic_auth = true
    
    return true
end
return M
